<?php

namespace WooCommerce_Product_Filter_Plugin\Project;

use WooCommerce_Product_Filter_Plugin\Structure;

class Post_Type extends Structure\Component {
	public function get_post_type() {
		return 'wcpf_project';
	}

	public function get_item_post_type() {
		return 'wcpf_item';
	}

	public function initial_properties() {
		$this->save_component_to_register( 'Project/Post_Type' );
	}

	public function attach_hooks( Structure\Hook_Manager $hook_manager ) {
		$hook_manager->add_action( 'init', 'on_init' );

		// Add action for when a post is trashed.
		add_action(
			'wp_trash_post',
			function ( $post_id ) {
				if ( $this->is_project_post_type( $post_id ) ) {
					$this->on_delete_post( $post_id );
				}
			}
		);

		add_action(
			'untrash_post',
			function ( $post_id ) {
				if ( $this->is_project_post_type( $post_id ) ) {
					$this->on_untrash_post( $post_id );
				}
			}
		);

		// Add action for when a post is deleted.
		add_action(
			'delete_post',
			function ( $post_id ) {
				if ( $this->is_project_post_type( $post_id ) ) {
					$this->on_delete_post( $post_id, true );
				}
			}
		);

		// Filter the post status when a post is untrashed.
		add_filter(
			'wp_untrash_post_status',
			function ( $new_status, $post_id, $previous_status ) {
				switch ( get_post_type( $post_id ) ) {
					case $this->get_item_post_type():
					case $this->get_post_type():
						return $previous_status;

					default:
						return $new_status;
				}
			},
			10,
			3
		);
	}

	/**
	 * Whether the given post ID is the project post type.
	 *
	 * @since x.x.x
	 *
	 * @param int $post_id The post ID.
	 *
	 * @return bool
	 */
	private function is_project_post_type( $post_id ) {
		return get_post_type( $post_id ) === $this->get_post_type();
	}

	public function on_init() {
		register_post_type(
			$this->get_post_type(),
			array(
				'public'              => false,
				'has_archive'         => false,
				'publicaly_queryable' => false,
				'show_in_menu'        => 'woocommerce',
				'show_in_admin_bar'   => false,
				'show_ui'             => true,
				'hierarchical'        => false,
				'supports'            => array(
					'author',
				),
				'labels'              => array(
					'name'                  => __( 'Filters', 'woocommerce-product-filters' ),
					'singular_name'         => __( 'Filter', 'woocommerce-product-filters' ),
					'add_new'               => __( 'Add New Filter', 'woocommerce-product-filters' ),
					'add_new_item'          => __( 'Add New Filter', 'woocommerce-product-filters' ),
					'edit_item'             => __( 'Edit Filter', 'woocommerce-product-filters' ),
					'new_item'              => __( 'New Filter', 'woocommerce-product-filters' ),
					'view_item'             => __( 'View Filter', 'woocommerce-product-filters' ),
					'search_items'          => __( 'Search Filters', 'woocommerce-product-filters' ),
					'item_trashed'          => __( 'Filter trashed', 'woocommerce-product-filters' ),
					'not_found'             => __( 'No filters found.', 'woocommerce-product-filters' ),
					'not_found_in_trash'    => __( 'No filters found in trash.', 'woocommerce-product-filters' ),
					'parent'                => __( 'Parent Filters', 'woocommerce-product-filters' ),
					'filter_items_list'     => __( 'Filter Filters', 'woocommerce-product-filters' ),
					'items_list_navigation' => __( 'Filters navigation', 'woocommerce-product-filters' ),
					'items_list'            => __( 'Filters list', 'woocommerce-product-filters' ),
				),
			)
		);

		add_filter(
			'bulk_post_updated_messages',
			function ( $bulk_messages, $bulk_counts ) {
				$bulk_messages[ $this->get_post_type() ] = array(
					/* translators: %s: Number of filters. */
					'updated'   => _n( '%s filter updated.', '%s filters updated.', $bulk_counts['updated'], 'woocommerce-product-filters' ),
					/* translators: %s: Number of filters. */
					'locked'    => _n( '%s filter not updated, somebody is editing it.', '%s filters not updated, somebody is editing them.', $bulk_counts['locked'], 'woocommerce-product-filters' ),
					/* translators: %s: Number of filters. */
					'deleted'   => _n( '%s filter permanently deleted.', '%s filters permanently deleted.', $bulk_counts['deleted'], 'woocommerce-product-filters' ),
					/* translators: %s: Number of filters. */
					'trashed'   => _n( '%s filter moved to the Trash.', '%s filters moved to the Trash.', $bulk_counts['trashed'], 'woocommerce-product-filters' ),
					/* translators: %s: Number of filters. */
					'untrashed' => _n( '%s filter restored from the Trash.', '%s filters restored from the Trash.', $bulk_counts['untrashed'], 'woocommerce-product-filters' ),
				);
				return $bulk_messages;
			},
			10,
			2
		);

		register_post_type(
			$this->get_item_post_type(),
			array(
				'public'       => false,
				'hierarchical' => false,
				'supports'     => array(),
				'labels'       => array(
					'name' => __( 'Filter Item', 'woocommerce-product-filters' ),
				),
			)
		);
	}

	/**
	 * Delete all entities associated with a project.
	 *
	 * @since x.x.x Trash entities when the project is trashed.
	 *
	 * @param int  $post_id   The project post ID.
	 * @param bool $permanent Whether the project is being permanently deleted or trashed. Defaults to false (trashed).
	 *
	 * @return void
	 */
	public function on_delete_post( $post_id, bool $permanent = false ) {
		$project = $this->get_project( $post_id );
		foreach ( $project->get_entity_list() as $id => $entity ) {
			if ( $id === $post_id ) {
				continue;
			}

			if ( $permanent ) {
				wp_delete_post( $id, true );
			} else {
				wp_trash_post( $id );
			}
		}
	}

	/**
	 * Retrieve a project object from the post ID.
	 *
	 * @since x.x.x
	 *
	 * @param int $post_id The post ID for the project.
	 *
	 * @return Project
	 */
	private function get_project( int $post_id ) {
		/** @var Project $project */
		$project = $this->get_component_builder()->build( Project::class );
		$project->load_project( $post_id );

		return $project;
	}

	/**
	 * Untrash items when the project is untrashed.
	 *
	 * @since x.x.x
	 *
	 * @param int $post_id The post ID being untrashed.
	 *
	 * @return void
	 */
	private function on_untrash_post( $post_id ) {
		$project = $this->get_project( $post_id );
		foreach ( $project->get_entity_list() as $id => $entity ) {
			if ( $id === $post_id ) {
				continue;
			}

			wp_untrash_post( $id );
		}
	}
}
